############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Ftp.py,v 1.18 2000/10/10 18:30:57 bazsi Exp $
##
## Author  : kisza
## Auditor :  
## Last version: 
## Notes:
##
############################################################################

import Zorp, Service, Connector, Proxy, Session, Stream, SockAddr, Plug
from Zorp import *

#
# Mode flag bitmap
#
rfc1545		=	0x80000000
rfc2228		=	0x40000000
rfc1579		=	0x20000000
rfc2389		=	0x10000000
rfc2428		=	0x08000000
rfc765		=	0x04000000
rfc775		=	0x02000000
rfc959		=	0x01000000
levelConnect	=	0x00800000
levelLogin	=	0x00400000
levelDataprot	=	0x00200000
levelCommandc	=	0x00100000
levelLogoff	=	0x00080000
levelTranspar	=	0x00040000
levelInform	=	0x00020000
levelMisc	=	0x00010000
levelDirectory	=	0x00008000
levelFileact	=	0x00004000
levelMail	=	0x00002000
levelUnknown	=	0x00001000
accessRead	=	0x00000800
accessWrite	=	0x00000400
accessCtrlsys	=	0x00000200
accessCtrlnet	=	0x00000100
accessUnkn	=	0x00000080
nameKnown	=	0x00000040
nameUnknown	=	0x00000020
nameReset	=	0x00000010
ccRead		=	0x00000008
ccWrite		=	0x00000004
ccControll	=	0x00000002
ccFeedback	=	0x00000001

#
# Predefinied modes
#
AnonymousRead	=       0x01BFFFFB
AllowAllSTD9	=	0x01FFFFFF

class FtpDataConnect:
        state = 0
        parms = [(), ()];
        listen = None
        connect = None
	listens = [None, None]
        ftp_proxy = None
        fds = [-1, -1]
        session = None

        def __init__(self, session, ftp_proxy):
        	self.session = session
        	self.ftp_proxy = ftp_proxy
        	
        def __del__(self):
        	self.resetState()

	def resetState(self):
		if self.listen != None:
	        	self.listen.destroy()
	        if self.connect != None:
	        	self.connect.destroy()
	        if self.listens[0] != None:
	       		self.listens[0].destroy()
	       	if self.listens[1] != None:
	        	self.listens[1].destroy()
		self.listens = [None, None]
		self.listen = None
		self.connect = None
		self.state = 0

	def prepare(self, ip, side):
		if (side < 0 or side > 1):
			return 0
			
		if self.listens[side] != None:
			self.listens[side].destroy()
		
		try:
#			self.listens[side] = \
#				Zorp.ListenPortRange(SockAddr.SockAddrInet(ip, 0), 
#						     self.ftp_proxy.data_port_min,
#						     self.ftp_proxy.data_port_max,
#						     self.acceptedCallback, 
#						     1)
			self.listens[side] = \
				Zorp.Listen(SockAddr.SockAddrInet(ip, 0), 
					    self.acceptedCallback, 
					    1)
			return self.listens[side].local.port
		except IOError:
			self.listens[side] = None
			
		return 0

	def start(self, fst_mode, fst_local, fst_remote,
	                snd_mode, snd_local, snd_remote):

                self.parms = [(fst_mode, fst_local, fst_remote),
                	      (snd_mode, snd_local, snd_remote)]
		self.state = 0
                self.processState()
                
        def processState(self):
		Zorp.debug(0, "data listener, state %d" % (self.state))
        	if self.state == 2:
        		# start proxy
			Zorp.debug(0, "stacking data proxy")
			ss = Session.StackedSession(self.session)
			ss.client_fd = self.fds[0]
			ss.client = Stream.Stream(ss.client_fd, ss.session_id + 'data client')
			ss.server_fd = self.fds[1]
			ss.server = Stream.Stream(ss.client_fd, ss.session_id + 'data server')
			self.ftp_proxy.data_proxy(ss)
        		return
        		
                parms = self.parms[self.state]
                if parms[0] == 'L':
			Zorp.debug(0, "data listener, standard")
			if self.listen != None:
				self.listen.destroy()
			try:
	                        self.listen = Zorp.Listen(parms[1], self.acceptedCallback, 1)
	                except IOError:
	                	self.listen = None
                        self.listen.start()
                elif parms[0] == 'C':
			Zorp.debug(0, "data connector")
			if self.connect != None:
				self.connect.destroy()
			try:
	                        self.connect = Zorp.Connect(parms[1], parms[2], self.connectedCallback)
			except IOError:
				self.connect = None
			
                elif parms[0] == 'A':
			Zorp.debug(0, "data listener, A")
                	self.listens[0].start()
              	elif parms[0] == 'B':
			Zorp.debug(0, "data listener, B")
              		self.listens[1].start()

        def acceptedCallback(self, client, fd):
		Zorp.debug(0, "data listener, acceptedCallback")
        	self.fds[self.state] = fd
                self.state = self.state + 1
                self.processState()

        def connectedCallback(self, fd):
		Zorp.debug(0, "data listener, connectedCallback")
        	self.fds[self.state] = fd
                self.state = self.state + 1
                self.processState()


# 
# Default FTP proxy - deny all
#
class FtpProxy(Proxy.Proxy):
	data_connect = None
	data_proxy = Plug.PlugProxy
	data_port_min = 4000;
	data_port_max = 6000;

	def __init__(self, session):
		Proxy.Proxy.__init__(self, "ftp", session)
		self.data_connect = FtpDataConnect(session, self)
		#, self.data_proxy_min, self.data_proxy_max)

	def user(self, dir, uname):
		#Zorp.debug(0, "FTPE: USER %s , flag %c" % (uname, dir))
		return Z_REJECT

	def password(self, dir, uname):
		#Zorp.debug(0, "FTPE: PASS , flag %c" % (dir))
		return Z_REJECT

	def nontransparent(self, dir, dst, user, passw):
		#Zorp.debug(0, "FTPE: nont %s %s %s, flag %c" % (dst, user, passw, dir))
		return Z_REJECT

	def userAUTH(self, dir, uname):
		#Zorp.debug(0, "FTPE: AUTH %s , flag %c" % (uname, dir))
		return Z_REJECT

	def quit(self, dir):
		#Zorp.debug(0, "FTPE: QUIT flag %c" % (dir))
		return Z_REJECT

	def bounceCheck(self, dir, realip, realport, reqip, reqport):
		#Zorp.debug(0, "FTPE: bounce attack detected")
		return Z_REJECT

        def portCommand(self, dir, hashv, comm, param, ip, port):
                #Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
                #Zorp.debug(0, "FTPE: %s:%d" % (ip, port))
                return Z_REJECT

        def pasvAnswer(self, dir, line, ip, port):
                #Zorp.debug(0, "FTPE: %s flag: %s" % (line, dir))
                #Zorp.debug(0, "FTPE: %s:%d" % (ip, port))
                return Z_REJECT

        def answer(self, dir, line):
                #Zorp.debug(0, "FTPE: %s flag: %s" % (line, dir))
                return Z_REJECT

	def transferCommand(self, dir, hashv, comm, param):
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def dataTransfer(self, dir, hashv, comm, param):
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def directoryChange(self, dir, hashv, comm, param):
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def directoryManipulation(self, dir, hashv, comm, param):
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def siteCommand(self, dir, hashv, comm, param):
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def mailCommand(self, dir, hashv, comm, param):
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def miscCommand(self, dir, hashv, comm, param):
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def generalCommand(self, dir, hashv, comm, param):
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def unknownCommand(self, dir, comm, param):
		#Zorp.debug(0, "FTPE: command: %s %s flag: %s" % ( comm, param, dir))
		return Z_REJECT

	def sendABOR(self, dir):
		#Zorp.debug(0, "FTPE: ABOR flag: %s" % ( dir))
		return Z_REJECT

	def prepareDATA(self, ip, side):
		return self.data_connect.prepare(ip, side)

	def startDATA(self, dir, fst_mode, fst_loc_addr, fst_loc_port, fst_rem_addr, fst_rem_port, snd_mode, snd_loc_addr, snd_loc_port, snd_rem_addr, snd_rem_port):
		#
		# mode-ok:
		#  L - listen
		#  C - connect
		#  A - 0. oldal elore foglalt porttal listen
		#  B - 1. oldal elore foglalt porttal listen
		#
		#Zorp.debug(0, "FTPE: Rem %s:%d - Loc %s:%d -%c" % (fst_rem_addr, fst_rem_port, fst_loc_addr, fst_loc_port, fst_mode))
		#Zorp.debug(0, "FTPE: Rem %s:%d - Loc %s:%d -%c" % (snd_rem_addr, snd_rem_port, snd_loc_addr, snd_loc_port, snd_mode))
		#Zorp.debug(0, "FTPE:  %s" % (dir))

		def convertSockAddr(addr, port):
			if addr == '':
				return None
			else:
				return SockAddr.SockAddrInet(addr, port)

		self.data_connect.start(fst_mode, 
					convertSockAddr(fst_loc_addr, fst_loc_port),
					convertSockAddr(fst_rem_addr, fst_rem_port),
					snd_mode,
					convertSockAddr(snd_loc_addr, snd_loc_port),
					convertSockAddr(snd_rem_addr, snd_rem_port))
		return Z_ACCEPT

#
#  Ftp proxy - allows anonymous download
#
class FtpProxyAnonymous(FtpProxy):

	def startUp(self):
		self.command_accept = AnonymousRead

	def config(self):
		self.command_accept = AnonymousRead
		
	def user(self, dir, uname):
		if  ( uname == "anonymous" or uname == "ftp" ):
			return Z_ACCEPT
		else:
			return Z_REJECT

	def password(self, dir, uname):
		return Z_ACCEPT

	def nontransparent(self, dir, dst, user, passw):
		if  ( uname == "anonymous" or uname == "ftp" ):
			return Z_ACCEPT
		else:
			return Z_REJECT

        def portCommand(self, dir, hashv, comm, param, ip, port):
                return Z_ACCEPT

        def pasvAnswer(self, dir, line, ip, port):
                return Z_ACCEPT

	def unknownCommand(self, dir, comm, param):
		return Z_ACCEPT

#
# Ftp proxy - R U kidding? Allow all
#
class FtpProxyAllow(FtpProxy):

	def startUp(self):
		self.command_accept = AllowAllSTD9

	def config(self):
		self.command_accept = AllowAllSTD9
		
        def portCommand(self, dir, hashv, comm, param, ip, port):
                return Z_ACCEPT

        def pasvAnswer(self, dir, line, ip, port):
                return Z_ACCEPT

	def unknownCommand(self, dir, comm, param):
		return Z_ACCEPT

